from autoarray.plot import plotters
from autoastro.plot import lensing_plotters


@lensing_plotters.set_include_and_sub_plotter
@plotters.set_labels
def subplot_image_and_mapper(
    image,
    mapper,
    image_positions=None,
    source_positions=None,
    critical_curves=None,
    caustics=None,
    image_pixel_indexes=None,
    source_pixel_indexes=None,
    include=plotters.Include(),
    sub_plotter=None,
):

    number_subplots = 2

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=1)

    sub_plotter.plot_array(
        array=image,
        mask=include.mask_from_grid(grid=mapper.grid),
        positions=image_positions,
        lines=critical_curves,
        include_origin=include.origin,
    )

    if image_pixel_indexes is not None:

        sub_plotter.index_scatterer.scatter_grid_indexes(
            grid=image.mask.geometry.masked_grid, indexes=image_pixel_indexes
        )

    if source_pixel_indexes is not None:

        indexes = mapper.image_pixel_indexes_from_source_pixel_indexes(
            source_pixel_indexes=source_pixel_indexes
        )

        sub_plotter.index_scatterer.scatter_grid_indexes(
            grid=image.mask.geometry.masked_grid, indexes=indexes
        )

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=2)

    sub_plotter.plot_mapper(
        mapper=mapper,
        positions=source_positions,
        caustics=caustics,
        image_pixel_indexes=image_pixel_indexes,
        source_pixel_indexes=source_pixel_indexes,
        include_origin=include.origin,
        include_grid=include.inversion_grid,
        include_pixelization_grid=include.inversion_pixelization_grid,
        include_border=include.inversion_border,
    )

    sub_plotter.output.subplot_to_figure()
    sub_plotter.figure.close()



import autolens as al
import autolens.plot as aplt
import os

import pytest

from os import path

from autofit import conf

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(name="ray_tracing_plotter_path")
def make_ray_tracing_plotter_setup():
    return "{}/../../test_files/plotting/ray_tracing/".format(
        os.path.dirname(os.path.realpath(__file__))
    )


@pytest.fixture(autouse=True)
def set_config_path():
    conf.instance = conf.Config(
        path.join(directory, "../test_files/plot"), path.join(directory, "output")
    )


def test__all_individual_plotters(
    tracer_x2_plane_7x7,
    sub_grid_7x7,
    mask_7x7,
    include_all,
    ray_tracing_plotter_path,
    plot_patch,
):
    aplt.tracer.profile_image(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        include=include_all,
        plotter=aplt.Plotter(
            output=aplt.Output(ray_tracing_plotter_path, format="png")
        ),
    )

    assert ray_tracing_plotter_path + "profile_image.png" in plot_patch.paths

    aplt.tracer.convergence(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        include=include_all,
        plotter=aplt.Plotter(
            output=aplt.Output(ray_tracing_plotter_path, format="png")
        ),
    )

    assert ray_tracing_plotter_path + "convergence.png" in plot_patch.paths

    aplt.tracer.potential(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        include=include_all,
        plotter=aplt.Plotter(
            output=aplt.Output(ray_tracing_plotter_path, format="png")
        ),
    )

    assert ray_tracing_plotter_path + "potential.png" in plot_patch.paths

    aplt.tracer.deflections_y(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        include=include_all,
        plotter=aplt.Plotter(
            output=aplt.Output(ray_tracing_plotter_path, format="png")
        ),
    )

    assert ray_tracing_plotter_path + "deflections_y.png" in plot_patch.paths

    aplt.tracer.deflections_x(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        include=include_all,
        plotter=aplt.Plotter(
            output=aplt.Output(ray_tracing_plotter_path, format="png")
        ),
    )

    assert ray_tracing_plotter_path + "deflections_x.png" in plot_patch.paths

    aplt.tracer.magnification(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        include=include_all,
        plotter=aplt.Plotter(
            output=aplt.Output(ray_tracing_plotter_path, format="png")
        ),
    )

    assert ray_tracing_plotter_path + "magnification.png" in plot_patch.paths

    tracer_x2_plane_7x7.planes[0].galaxies[0].hyper_galaxy = al.HyperGalaxy()
    tracer_x2_plane_7x7.planes[0].galaxies[0].hyper_model_image = al.array.ones(
        shape_2d=(7, 7), pixel_scales=0.1
    )
    tracer_x2_plane_7x7.planes[0].galaxies[0].hyper_galaxy_image = al.array.ones(
        shape_2d=(7, 7), pixel_scales=0.1
    )

    aplt.tracer.contribution_map(
        tracer=tracer_x2_plane_7x7,
        mask=mask_7x7,
        include=include_all,
        plotter=aplt.Plotter(
            output=aplt.Output(ray_tracing_plotter_path, format="png")
        ),
    )

    assert ray_tracing_plotter_path + "contribution_map.png" in plot_patch.paths


def test__tracer_sub_plot_output(
    tracer_x2_plane_7x7, sub_grid_7x7, include_all, ray_tracing_plotter_path, plot_patch
):
    aplt.tracer.subplot_tracer(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        include=include_all,
        sub_plotter=aplt.SubPlotter(
            output=aplt.Output(ray_tracing_plotter_path, format="png")
        ),
    )

    assert ray_tracing_plotter_path + "subplot_tracer.png" in plot_patch.paths


def test__tracer_individuals__dependent_on_input(
    tracer_x2_plane_7x7, sub_grid_7x7, include_all, ray_tracing_plotter_path, plot_patch
):
    aplt.tracer.individual(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        plot_profile_image=True,
        plot_source_plane=True,
        plot_potential=True,
        plot_magnification=True,
        include=include_all,
        plotter=aplt.Plotter(
            output=aplt.Output(ray_tracing_plotter_path, format="png")
        ),
    )

    assert ray_tracing_plotter_path + "profile_image.png" in plot_patch.paths

    assert ray_tracing_plotter_path + "source_plane.png" in plot_patch.paths

    assert ray_tracing_plotter_path + "convergence.png" not in plot_patch.paths

    assert ray_tracing_plotter_path + "potential.png" in plot_patch.paths

    assert ray_tracing_plotter_path + "deflections_y.png" not in plot_patch.paths

    assert ray_tracing_plotter_path + "deflections_x.png" not in plot_patch.paths

    assert ray_tracing_plotter_path + "magnification.png" in plot_patch.paths



import os

import pytest
from matplotlib import pyplot

import autofit as af


@pytest.fixture(name="general_config", autouse=True)
def make_general_config():
    directory = os.path.dirname(os.path.realpath(__file__))
    config_path = f"{directory}/integration/config"
    plotting_config_path = f"{config_path}/plotting/"
    af.conf.instance.general = af.conf.NamedConfig(plotting_config_path + "general.ini")
    af.conf.instance.config_path = config_path


class PlotPatch:
    def __init__(self):
        self.paths = []

    def __call__(self, path, *args, **kwargs):
        self.paths.append(path)


@pytest.fixture(name="plot_patch")
def make_plot_patch(monkeypatch):
    plot_patch = PlotPatch()
    monkeypatch.setattr(pyplot, "savefig", plot_patch)
    return plot_patch



from autoarray.exc import InversionException, GridException
from autofit.exc import FitException
from autolens.fit import fit
from autolens.pipeline import visualizer
from autolens.pipeline.phase.dataset import analysis as analysis_dataset


class Analysis(analysis_dataset.Analysis):
    def __init__(self, masked_imaging, cosmology, image_path=None, results=None):

        super(Analysis, self).__init__(cosmology=cosmology, results=results)

        self.visualizer = visualizer.PhaseImagingVisualizer(
            masked_dataset=masked_imaging, image_path=image_path, results=results
        )

        self.masked_dataset = masked_imaging

    @property
    def masked_imaging(self):
        return self.masked_dataset

    def fit(self, instance):
        """
        Determine the fit of a lens galaxy and source galaxy to the masked_imaging in this lens.

        Parameters
        ----------
        instance
            A model instance with attributes

        Returns
        -------
        fit : Fit
            A fractional value indicating how well this model fit and the model masked_imaging itself
        """

        self.associate_hyper_images(instance=instance)
        tracer = self.tracer_for_instance(instance=instance)

        self.masked_dataset.check_positions_trace_within_threshold_via_tracer(
            tracer=tracer
        )

        self.masked_dataset.check_inversion_pixels_are_below_limit_via_tracer(
            tracer=tracer
        )

        hyper_image_sky = self.hyper_image_sky_for_instance(instance=instance)

        hyper_background_noise = self.hyper_background_noise_for_instance(
            instance=instance
        )

        try:
            fit = self.masked_imaging_fit_for_tracer(
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

            return fit.figure_of_merit
        except InversionException or GridException as e:
            raise FitException from e

    def masked_imaging_fit_for_tracer(
        self, tracer, hyper_image_sky, hyper_background_noise
    ):

        return fit.ImagingFit(
            masked_imaging=self.masked_dataset,
            tracer=tracer,
            hyper_image_sky=hyper_image_sky,
            hyper_background_noise=hyper_background_noise,
        )

    def visualize(self, instance, during_analysis):
        instance = self.associate_hyper_images(instance=instance)
        tracer = self.tracer_for_instance(instance=instance)
        hyper_image_sky = self.hyper_image_sky_for_instance(instance=instance)
        hyper_background_noise = self.hyper_background_noise_for_instance(
            instance=instance
        )

        fit = self.masked_imaging_fit_for_tracer(
            tracer=tracer,
            hyper_image_sky=hyper_image_sky,
            hyper_background_noise=hyper_background_noise,
        )

        if tracer.has_mass_profile:

            visualizer = self.visualizer.new_visualizer_with_preloaded_critical_curves_and_caustics(
                preloaded_critical_curves=tracer.critical_curves,
                preloaded_caustics=tracer.caustics,
            )

        else:

            visualizer = self.visualizer

        visualizer.visualize_ray_tracing(
            tracer=fit.tracer, during_analysis=during_analysis
        )
        visualizer.visualize_fit(fit=fit, during_analysis=during_analysis)



import autolens as al
import autolens.plot as aplt

from test_autolens.simulate.imaging import simulate_util

# In this tutorial, we'll introduce a new pixelization, called an adaptive-pixelization. This pixelization doesn't use
# uniform grid of rectangular pixels, but instead uses ir'Voronoi' pixels. So, why would we want to do that?
# Lets take another look at the rectangular grid, and think about its weakness.

# Lets quickly remind ourselves of the image, and the 3.0" circular mask we'll use to mask it.
imaging = simulate_util.load_test_imaging(
    data_type="lens_mass__source_cuspy", data_resolution="hst"
)
mask = al.mask.circular(
    shape_2d=imaging.shape_2d,
    pixel_scales=imaging.pixel_scales,
    radius=3.0,
    centre=(0.0, 0.0),
    sub_size=4,
)

lens_galaxy = al.Galaxy(
    redshift=0.5, mass=al.mp.SphericalIsothermal(centre=(0.1, 0.2), einstein_radius=1.5)
)

source_galaxy = al.Galaxy(
    redshift=1.0,
    pixelization=al.pix.VoronoiMagnification(shape=(20, 20)),
    regularization=al.reg.Constant(coefficient=1.0),
)

masked_imaging = al.masked.imaging(imaging=imaging, mask=mask)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])
fit = al.fit(masked_dataset=masked_imaging, tracer=tracer)

print(fit.inversion.reconstruction)
print(fit.inversion.mapper)
print(fit.inversion.mapper.pixelization_grid)

print(fit.inversion.interpolated_reconstruction_from_shape_2d())

# aplt.inversion.interpolated_reconstruction(inversion=fit.inversion)

# aplt.fit_imaging.subplot_of_plane(
#     fit=fit, plane_index=1,
# )
#
# aplt.fit_imaging.subplot_of_plane(
#     fit=fit,
#     plane_index=1,
# )

aplt.fit_imaging.subplot_fit_imaging(fit=fit)
aplt.inversion.subplot_inversion(inversion=fit.inversion, caustics=tracer.caustics)



import os
from os import path

import pytest

import autofit as af
import autolens as al
from test_autolens.mock import mock_pipeline

pytestmark = pytest.mark.filterwarnings(
    "ignore:Using a non-tuple sequence for multidimensional indexing is deprecated; use `arr[tuple(seq)]` instead of "
    "`arr[seq]`. In the future this will be interpreted as an arrays index, `arr[np.arrays(seq)]`, which will result "
    "either in an error or a different result."
)

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(scope="session", autouse=True)
def do_something():
    print("{}/../test_files/config/".format(directory))

    af.conf.instance = af.conf.Config("{}/../../test_files/config/".format(directory))


def clean_images():
    try:
        os.remove("{}/source_lens_phase/source_image_0.fits".format(directory))
        os.remove("{}/source_lens_phase/lens_image_0.fits".format(directory))
        os.remove("{}/source_lens_phase/model_image_0.fits".format(directory))
    except FileNotFoundError:
        pass
    af.conf.instance.dataset_path = directory


class TestPhase:

    # TODO : These tests have both turned to models?

    def test__set_instances(self, phase_dataset_7x7):
        phase_dataset_7x7.galaxies = [al.Galaxy(redshift=0.5)]
        assert phase_dataset_7x7.model.galaxies == [al.Galaxy(redshift=0.5)]

    def test__set_models(self, phase_dataset_7x7):
        phase_dataset_7x7.galaxies = [al.GalaxyModel(redshift=0.5)]
        assert phase_dataset_7x7.model.galaxies == [al.GalaxyModel(redshift=0.5)]

    def test__promise_attrbutes(self):
        phase = al.PhaseDataset(
            galaxies=dict(
                lens=al.GalaxyModel(
                    redshift=0.5,
                    mass=al.mp.EllipticalIsothermal,
                    shear=al.mp.ExternalShear,
                ),
                source=al.GalaxyModel(redshift=1.0, light=al.lp.EllipticalSersic),
            ),
            optimizer_class=mock_pipeline.MockNLO,
            phase_tag="",
            phase_name="test_phase",
        )

        print(hasattr(af.last.result.instance.galaxies.lens, "mas2s"))

    def test__customize(
        self, results_7x7, results_collection_7x7, imaging_7x7, mask_7x7
    ):
        class MyPlanePhaseAnd(al.PhaseImaging):
            def customize_priors(self, results):
                self.galaxies = results.last.instance.galaxies

        galaxy = al.Galaxy(redshift=0.5)
        galaxy_model = al.GalaxyModel(redshift=0.5)

        setattr(results_7x7.instance, "galaxies", [galaxy])
        setattr(results_7x7.model, "galaxies", [galaxy_model])

        phase_dataset_7x7 = MyPlanePhaseAnd(
            phase_name="test_phase", optimizer_class=mock_pipeline.MockNLO
        )

        phase_dataset_7x7.make_analysis(
            dataset=imaging_7x7, mask=mask_7x7, results=results_collection_7x7
        )
        phase_dataset_7x7.customize_priors(results_collection_7x7)

        assert phase_dataset_7x7.galaxies == [galaxy]

        class MyPlanePhaseAnd(al.PhaseImaging):
            def customize_priors(self, results):
                self.galaxies = results.last.model.galaxies

        galaxy = al.Galaxy(redshift=0.5)
        galaxy_model = al.GalaxyModel(redshift=0.5)

        setattr(results_7x7.instance, "galaxies", [galaxy])
        setattr(results_7x7.model, "galaxies", [galaxy_model])

        phase_dataset_7x7 = MyPlanePhaseAnd(
            phase_name="test_phase", optimizer_class=mock_pipeline.MockNLO
        )

        phase_dataset_7x7.make_analysis(
            dataset=imaging_7x7, mask=mask_7x7, results=results_collection_7x7
        )
        phase_dataset_7x7.customize_priors(results_collection_7x7)

        assert phase_dataset_7x7.galaxies == [galaxy_model]

    def test__duplication(self):
        phase_dataset_7x7 = al.PhaseImaging(
            phase_name="test_phase",
            galaxies=dict(
                lens=al.GalaxyModel(redshift=0.5), source=al.GalaxyModel(redshift=1.0)
            ),
        )

        al.PhaseImaging(phase_name="test_phase")

        assert phase_dataset_7x7.galaxies is not None

    def test__uses_pixelization_preload_grids_if_possible(self, imaging_7x7, mask_7x7):
        phase_dataset_7x7 = al.PhaseImaging(phase_name="test_phase")

        analysis = phase_dataset_7x7.make_analysis(dataset=imaging_7x7, mask=mask_7x7)

        assert analysis.masked_dataset.preload_sparse_grids_of_planes is None

    def test__phase_can_receive_list_of_galaxy_models(self):
        phase_dataset_7x7 = al.PhaseImaging(
            galaxies=dict(
                lens=al.GalaxyModel(
                    sersic=al.lp.EllipticalSersic,
                    sis=al.mp.SphericalIsothermal,
                    redshift=al.Redshift,
                ),
                lens1=al.GalaxyModel(
                    sis=al.mp.SphericalIsothermal, redshift=al.Redshift
                ),
            ),
            optimizer_class=af.MultiNest,
            phase_name="test_phase",
        )

        for item in phase_dataset_7x7.model.path_priors_tuples:
            print(item)

        sersic = phase_dataset_7x7.model.galaxies[0].sersic
        sis = phase_dataset_7x7.model.galaxies[0].sis
        lens_1_sis = phase_dataset_7x7.model.galaxies[1].sis

        arguments = {
            sersic.centre[0]: 0.2,
            sersic.centre[1]: 0.2,
            sersic.axis_ratio: 0.0,
            sersic.phi: 0.1,
            sersic.effective_radius.priors[0]: 0.2,
            sersic.sersic_index: 0.6,
            sersic.intensity.priors[0]: 0.6,
            sis.centre[0]: 0.1,
            sis.centre[1]: 0.2,
            sis.einstein_radius.priors[0]: 0.3,
            phase_dataset_7x7.model.galaxies[0].redshift.priors[0]: 0.4,
            lens_1_sis.centre[0]: 0.6,
            lens_1_sis.centre[1]: 0.5,
            lens_1_sis.einstein_radius.priors[0]: 0.7,
            phase_dataset_7x7.model.galaxies[1].redshift.priors[0]: 0.8,
        }

        instance = phase_dataset_7x7.model.instance_for_arguments(arguments=arguments)

        assert instance.galaxies[0].sersic.centre[0] == 0.2
        assert instance.galaxies[0].sis.centre[0] == 0.1
        assert instance.galaxies[0].sis.centre[1] == 0.2
        assert instance.galaxies[0].sis.einstein_radius == 0.3
        assert instance.galaxies[0].redshift == 0.4
        assert instance.galaxies[1].sis.centre[0] == 0.6
        assert instance.galaxies[1].sis.centre[1] == 0.5
        assert instance.galaxies[1].sis.einstein_radius == 0.7
        assert instance.galaxies[1].redshift == 0.8

        class LensPlanePhase2(al.PhaseImaging):
            # noinspection PyUnusedLocal
            def pass_models(self, results):
                self.galaxies[0].sis.einstein_radius = 10.0

        phase_dataset_7x7 = LensPlanePhase2(
            galaxies=dict(
                lens=al.GalaxyModel(
                    sersic=al.lp.EllipticalSersic,
                    sis=al.mp.SphericalIsothermal,
                    redshift=al.Redshift,
                ),
                lens1=al.GalaxyModel(
                    sis=al.mp.SphericalIsothermal, redshift=al.Redshift
                ),
            ),
            optimizer_class=af.MultiNest,
            phase_name="test_phase",
        )

        # noinspection PyTypeChecker
        phase_dataset_7x7.pass_models(None)

        sersic = phase_dataset_7x7.model.galaxies[0].sersic
        sis = phase_dataset_7x7.model.galaxies[0].sis
        lens_1_sis = phase_dataset_7x7.model.galaxies[1].sis

        arguments = {
            sersic.centre[0]: 0.01,
            sersic.centre[1]: 0.2,
            sersic.axis_ratio: 0.0,
            sersic.phi: 0.1,
            sersic.effective_radius.priors[0]: 0.2,
            sersic.sersic_index: 0.6,
            sersic.intensity.priors[0]: 0.6,
            sis.centre[0]: 0.1,
            sis.centre[1]: 0.2,
            phase_dataset_7x7.model.galaxies[0].redshift.priors[0]: 0.4,
            lens_1_sis.centre[0]: 0.6,
            lens_1_sis.centre[1]: 0.5,
            lens_1_sis.einstein_radius.priors[0]: 0.7,
            phase_dataset_7x7.model.galaxies[1].redshift.priors[0]: 0.8,
        }

        instance = phase_dataset_7x7.model.instance_for_arguments(arguments)

        assert instance.galaxies[0].sersic.centre[0] == 0.01
        assert instance.galaxies[0].sis.centre[0] == 0.1
        assert instance.galaxies[0].sis.centre[1] == 0.2
        assert instance.galaxies[0].sis.einstein_radius == 10.0
        assert instance.galaxies[0].redshift == 0.4
        assert instance.galaxies[1].sis.centre[0] == 0.6
        assert instance.galaxies[1].sis.centre[1] == 0.5
        assert instance.galaxies[1].sis.einstein_radius == 0.7
        assert instance.galaxies[1].redshift == 0.8


class TestResult:
    def test__results_of_phase_are_available_as_properties(self, imaging_7x7, mask_7x7):
        clean_images()

        phase_dataset_7x7 = al.PhaseImaging(
            optimizer_class=mock_pipeline.MockNLO,
            galaxies=[
                al.Galaxy(redshift=0.5, light=al.lp.EllipticalSersic(intensity=1.0))
            ],
            phase_name="test_phase_2",
        )

        result = phase_dataset_7x7.run(dataset=imaging_7x7, mask=mask_7x7)

        assert isinstance(result, al.AbstractPhase.Result)

    def test__most_likely_tracer_available_as_result(self, imaging_7x7, mask_7x7):

        phase_dataset_7x7 = al.PhaseImaging(
            optimizer_class=mock_pipeline.MockNLO,
            galaxies=dict(
                lens=al.Galaxy(
                    redshift=0.5, light=al.lp.EllipticalSersic(intensity=1.0)
                ),
                source=al.Galaxy(
                    redshift=1.0, light=al.lp.EllipticalCoreSersic(intensity=2.0)
                ),
            ),
            phase_name="test_phase_2",
        )

        result = phase_dataset_7x7.run(dataset=imaging_7x7, mask=mask_7x7)

        assert isinstance(result.most_likely_tracer, al.Tracer)
        assert result.most_likely_tracer.galaxies[0].light.intensity == 1.0
        assert result.most_likely_tracer.galaxies[1].light.intensity == 2.0


class TestPhasePickle:

    # noinspection PyTypeChecker
    def test_assertion_failure(self, imaging_7x7, mask_7x7):
        def make_analysis(*args, **kwargs):
            return mock_pipeline.GalaxiesMockAnalysis(1, 1)

        phase_dataset_7x7 = al.PhaseImaging(
            phase_name="phase_name",
            optimizer_class=mock_pipeline.MockNLO,
            galaxies=dict(
                lens=al.Galaxy(light=al.lp.EllipticalLightProfile, redshift=1)
            ),
        )

        phase_dataset_7x7.make_analysis = make_analysis
        result = phase_dataset_7x7.run(
            dataset=imaging_7x7, results=None, mask=None, positions=None
        )
        assert result is not None

        phase_dataset_7x7 = al.PhaseImaging(
            phase_name="phase_name",
            optimizer_class=mock_pipeline.MockNLO,
            galaxies=dict(
                lens=al.Galaxy(light=al.lp.EllipticalLightProfile, redshift=1)
            ),
        )

        phase_dataset_7x7.make_analysis = make_analysis
        result = phase_dataset_7x7.run(
            dataset=imaging_7x7, results=None, mask=None, positions=None
        )
        assert result is not None

        class CustomPhase(al.PhaseImaging):
            def customize_priors(self, results):
                self.galaxies.lens.light = al.lp.EllipticalLightProfile()

        phase_dataset_7x7 = CustomPhase(
            phase_name="phase_name",
            optimizer_class=mock_pipeline.MockNLO,
            galaxies=dict(
                lens=al.Galaxy(light=al.lp.EllipticalLightProfile, redshift=1)
            ),
        )
        phase_dataset_7x7.make_analysis = make_analysis

        # with pytest.raises(af.exc.PipelineException):
        #     phase_dataset_7x7.run(data_type=imaging_7x7, results=None, mask=None, positions=None)



from test_autolens.integration.tests.imaging.lens_only import lens_x2_light
from test_autolens.integration.tests.imaging.runner import run_a_mock


class TestCase:
    def _test__lens_x2_light(self):
        run_a_mock(lens_x2_light)



import numpy as np

import autofit as af
import autolens as al


class GalaxiesMockAnalysis:
    def __init__(self, number_galaxies, value):
        self.number_galaxies = number_galaxies
        self.value = value

    # noinspection PyUnusedLocal
    def galaxy_images_for_model(self, model):
        return self.number_galaxies * [np.array([self.value])]

    def fit(self, instance):
        return 1


class MockResults:
    def __init__(
        self,
        model_image=None,
        mask=None,
        galaxy_images=(),
        instance=None,
        analysis=None,
        optimizer=None,
        pixelization=None,
    ):
        self.model_image = model_image
        self.unmasked_model_image = model_image
        self.mask_2d = mask
        self.galaxy_images = galaxy_images
        self.instance = instance or af.ModelInstance()
        self.model = af.ModelMapper()
        self.analysis = analysis
        self.optimizer = optimizer
        self.pixelization = pixelization
        self.hyper_combined = MockHyperCombinedPhase()

    @property
    def path_galaxy_tuples(self) -> [(str, al.Galaxy)]:
        """
        Tuples associating the names of galaxies with instances from the best fit
        """
        return [(("g0",), al.Galaxy(redshift=0.5)), (("g1",), al.Galaxy(redshift=1.0))]

    @property
    def path_galaxy_tuples_with_index(self) -> [(str, al.Galaxy)]:
        """
        Tuples associating the names of galaxies with instances from the best fit
        """
        return [
            (0, ("g0",), al.Galaxy(redshift=0.5)),
            (1, ("g1",), al.Galaxy(redshift=1.0)),
        ]

    @property
    def image_galaxy_dict(self) -> {str: al.Galaxy}:
        """
        A dictionary associating galaxy names with model images of those galaxies
        """
        return {
            galaxy_path: self.galaxy_images[i]
            for i, galaxy_path, galaxy in self.path_galaxy_tuples_with_index
        }

    @property
    def hyper_galaxy_image_path_dict(self):
        """
        A dictionary associating 1D hyper_galaxies galaxy images with their names.
        """

        hyper_minimum_percent = af.conf.instance.general.get(
            "hyper", "hyper_minimum_percent", float
        )

        hyper_galaxy_image_path_dict = {}

        for path, galaxy in self.path_galaxy_tuples:
            galaxy_image = self.image_galaxy_dict[path]

            minimum_galaxy_value = hyper_minimum_percent * max(galaxy_image)

            galaxy_image[galaxy_image < minimum_galaxy_value] = minimum_galaxy_value

            hyper_galaxy_image_path_dict[path] = galaxy_image

        return hyper_galaxy_image_path_dict

    @property
    def hyper_model_image(self):

        hyper_model_image = al.masked.array.zeros(mask=self.mask_2d)

        for path, galaxy in self.path_galaxy_tuples:
            hyper_model_image += self.hyper_galaxy_image_path_dict[path]

        return hyper_model_image


class MockResult:
    def __init__(self, instance, figure_of_merit, model=None):
        self.instance = instance
        self.figure_of_merit = figure_of_merit
        self.model = model
        self.previous_model = model
        self.gaussian_tuples = None
        self.mask_2d = None
        self.positions = None


class MockHyperCombinedPhase:
    def __init__(self):
        pass

    @property
    def most_likely_pixelization_grids_of_planes(self):
        return 1


class MockNLO(af.NonLinearOptimizer):
    def fit(self, analysis, model):
        class Fitness:
            def __init__(self, instance_from_physical_vector):
                self.result = None
                self.instance_from_physical_vector = instance_from_physical_vector

            def __call__(self, vector):
                instance = self.instance_from_physical_vector(vector)

                likelihood = analysis.fit(instance)
                self.result = MockResult(instance, likelihood)

                # Return Chi squared
                return -2 * likelihood

        fitness_function = Fitness(model.instance_from_physical_vector)
        fitness_function(model.prior_count * [0.8])

        return fitness_function.result



import autolens as al


class TestPipelineGeneralSettings:
    def test__hyper_galaxies_tag(self):

        general = al.setup.General(hyper_galaxies=False)
        assert general.hyper_galaxies_tag == ""

        general = al.setup.General(hyper_galaxies=True)
        assert general.hyper_galaxies_tag == "_galaxies"

    def test__hyper_image_sky_tag(self):
        general = al.setup.General(hyper_image_sky=False)
        assert general.hyper_galaxies_tag == ""

        general = al.setup.General(hyper_image_sky=True)
        assert general.hyper_image_sky_tag == "_bg_sky"

    def test__hyper_background_noise_tag(self):
        general = al.setup.General(hyper_background_noise=False)
        assert general.hyper_galaxies_tag == ""

        general = al.setup.General(hyper_background_noise=True)
        assert general.hyper_background_noise_tag == "_bg_noise"

    def test__tag(self):

        general = al.setup.General(
            hyper_galaxies=True, hyper_image_sky=True, hyper_background_noise=True
        )

        assert general.tag == "general__hyper_galaxies_bg_sky_bg_noise"

        general = al.setup.General(hyper_galaxies=True, hyper_background_noise=True)

        assert general.tag == "general__hyper_galaxies_bg_noise"


class TestPipelineSourceSettings:
    def test__pixelization_tag(self):
        source = al.setup.Source(pixelization=None)
        assert source.pixelization_tag == ""
        source = al.setup.Source(pixelization=al.pix.Rectangular)
        assert source.pixelization_tag == "pix_rect"
        source = al.setup.Source(pixelization=al.pix.VoronoiBrightnessImage)
        assert source.pixelization_tag == "pix_voro_image"

    def test__regularization_tag(self):
        source = al.setup.Source(regularization=None)
        assert source.regularization_tag == ""
        source = al.setup.Source(regularization=al.reg.Constant)
        assert source.regularization_tag == "__reg_const"
        source = al.setup.Source(regularization=al.reg.AdaptiveBrightness)
        assert source.regularization_tag == "__reg_adapt_bright"

    def test__lens_light_centre_tag(self):

        source = al.setup.Source(lens_light_centre=None)
        assert source.lens_light_centre_tag == ""
        source = al.setup.Source(lens_light_centre=(2.0, 2.0))
        assert source.lens_light_centre_tag == "__lens_light_centre_(2.00,2.00)"
        source = al.setup.Source(lens_light_centre=(3.0, 4.0))
        assert source.lens_light_centre_tag == "__lens_light_centre_(3.00,4.00)"
        source = al.setup.Source(lens_light_centre=(3.027, 4.033))
        assert source.lens_light_centre_tag == "__lens_light_centre_(3.03,4.03)"

    def test__lens_mass_centre_tag(self):

        source = al.setup.Source(lens_mass_centre=None)
        assert source.lens_mass_centre_tag == ""
        source = al.setup.Source(lens_mass_centre=(2.0, 2.0))
        assert source.lens_mass_centre_tag == "__lens_mass_centre_(2.00,2.00)"
        source = al.setup.Source(lens_mass_centre=(3.0, 4.0))
        assert source.lens_mass_centre_tag == "__lens_mass_centre_(3.00,4.00)"
        source = al.setup.Source(lens_mass_centre=(3.027, 4.033))
        assert source.lens_mass_centre_tag == "__lens_mass_centre_(3.03,4.03)"

    def test__align_light_mass_centre_tag__is_empty_sting_if_both_lens_light_and_mass_centres_input(
        self
    ):
        source = al.setup.Source(align_light_mass_centre=False)
        assert source.align_light_mass_centre_tag == ""
        source = al.setup.Source(align_light_mass_centre=True)
        assert source.align_light_mass_centre_tag == "__align_light_mass_centre"
        source = al.setup.Source(
            lens_light_centre=(0.0, 0.0),
            lens_mass_centre=(1.0, 1.0),
            align_light_mass_centre=True,
        )
        assert source.align_light_mass_centre_tag == ""

    def test__lens_light_bulge_only_tag(self):
        source = al.setup.Source(lens_light_bulge_only=False)
        assert source.lens_light_bulge_only_tag == ""
        source = al.setup.Source(lens_light_bulge_only=True)
        assert source.lens_light_bulge_only_tag == "__bulge_only"

    def test__no_shear_tag(self):
        source = al.setup.Source(no_shear=False)
        assert source.no_shear_tag == "__with_shear"

        source = al.setup.Source(no_shear=True)
        assert source.no_shear_tag == "__no_shear"

    def test__fix_lens_light_tag(self):
        source = al.setup.Source(fix_lens_light=False)
        assert source.fix_lens_light_tag == ""
        source = al.setup.Source(fix_lens_light=True)
        assert source.fix_lens_light_tag == "__fix_lens_light"

    def test__number_of_gaussians_tag(self):
        source = al.setup.Source()
        assert source.number_of_gaussians_tag == ""
        source = al.setup.Source(number_of_gaussians=1)
        assert source.number_of_gaussians_tag == "__gaussians_x1"
        source = al.setup.Source(number_of_gaussians=2)
        assert source.number_of_gaussians_tag == "__gaussians_x2"

    def test__tag(self):

        source = al.setup.Source(
            pixelization=al.pix.Rectangular,
            regularization=al.reg.Constant,
            lens_light_centre=(1.0, 2.0),
            lens_mass_centre=(3.0, 4.0),
            align_light_mass_centre=False,
            no_shear=True,
            fix_lens_light=True,
        )

        source.type_tag = source.inversion_tag

        assert (
            source.tag
            == "source__pix_rect__reg_const__no_shear__lens_light_centre_(1.00,2.00)__lens_mass_centre_(3.00,4.00)__fix_lens_light"
        )

        assert source.tag_beginner == "source__pix_rect__reg_const"

        source = al.setup.Source(
            pixelization=al.pix.Rectangular,
            regularization=al.reg.Constant,
            align_light_mass_centre=True,
            number_of_gaussians=1,
            fix_lens_light=True,
            lens_light_bulge_only=True,
        )

        source.type_tag = "test"

        assert (
            source.tag
            == "source__test__gaussians_x1__with_shear__align_light_mass_centre__bulge_only__fix_lens_light"
        )

    def test__tag_beginner(self):

        source = al.setup.Source(
            pixelization=al.pix.Rectangular, regularization=al.reg.Constant
        )

        assert source.tag_beginner_no_inversion == "source"

        assert source.tag_beginner == "source__pix_rect__reg_const"


class TestPipelineLightSettings:
    def test__align_bulge_disk_tags(self):

        light = al.setup.Light(align_bulge_disk_centre=False)
        assert light.align_bulge_disk_centre_tag == ""
        light = al.setup.Light(align_bulge_disk_centre=True)
        assert light.align_bulge_disk_centre_tag == "_centre"

        light = al.setup.Light(align_bulge_disk_axis_ratio=False)
        assert light.align_bulge_disk_axis_ratio_tag == ""
        light = al.setup.Light(align_bulge_disk_axis_ratio=True)
        assert light.align_bulge_disk_axis_ratio_tag == "_axis_ratio"

        light = al.setup.Light(align_bulge_disk_phi=False)
        assert light.align_bulge_disk_phi_tag == ""
        light = al.setup.Light(align_bulge_disk_phi=True)
        assert light.align_bulge_disk_phi_tag == "_phi"

    def test__bulge_disk_tag(self):
        light = al.setup.Light(
            align_bulge_disk_centre=False,
            align_bulge_disk_axis_ratio=False,
            align_bulge_disk_phi=False,
        )
        assert light.align_bulge_disk_tag == ""

        light = al.setup.Light(
            align_bulge_disk_centre=True,
            align_bulge_disk_axis_ratio=False,
            align_bulge_disk_phi=False,
        )
        assert light.align_bulge_disk_tag == "__align_bulge_disk_centre"

        light = al.setup.Light(
            align_bulge_disk_centre=True,
            align_bulge_disk_axis_ratio=False,
            align_bulge_disk_phi=True,
        )
        assert light.align_bulge_disk_tag == "__align_bulge_disk_centre_phi"

        light = al.setup.Light(
            align_bulge_disk_centre=True,
            align_bulge_disk_axis_ratio=True,
            align_bulge_disk_phi=True,
        )
        assert light.align_bulge_disk_tag == "__align_bulge_disk_centre_axis_ratio_phi"

    def test__disk_as_sersic_tag(self):
        light = al.setup.Light(disk_as_sersic=False)
        assert light.disk_as_sersic_tag == "__disk_exp"
        light = al.setup.Light(disk_as_sersic=True)
        assert light.disk_as_sersic_tag == "__disk_sersic"

    def test__number_of_gaussians_tag(self):
        source = al.setup.Source()
        assert source.number_of_gaussians_tag == ""
        source = al.setup.Source(number_of_gaussians=1)
        assert source.number_of_gaussians_tag == "__gaussians_x1"
        source = al.setup.Source(number_of_gaussians=2)
        assert source.number_of_gaussians_tag == "__gaussians_x2"

    def test__tag(self):

        light = al.setup.Light(align_bulge_disk_phi=True)
        light.type_tag = ""

        assert light.tag == "light____align_bulge_disk_phi__disk_exp"

        light = al.setup.Light(
            align_bulge_disk_centre=True,
            align_bulge_disk_axis_ratio=True,
            disk_as_sersic=True,
        )

        light.type_tag = "lol"

        assert (
            light.tag == "light__lol__align_bulge_disk_centre_axis_ratio__disk_sersic"
        )

        light = al.setup.Light(
            align_bulge_disk_centre=True,
            align_bulge_disk_axis_ratio=True,
            disk_as_sersic=True,
            number_of_gaussians=2,
        )
        light.type_tag = "test"

        assert light.tag == "light__test__gaussians_x2"

    def test__tag_beginner(self):

        light = al.setup.Light(align_bulge_disk_phi=True)

        assert light.tag_beginner == "light__align_bulge_disk_phi__disk_exp"

        light = al.setup.Light(
            align_bulge_disk_centre=True,
            align_bulge_disk_axis_ratio=True,
            disk_as_sersic=True,
        )

        assert (
            light.tag_beginner
            == "light__align_bulge_disk_centre_axis_ratio__disk_sersic"
        )

        light = al.setup.Light(
            align_bulge_disk_centre=True,
            align_bulge_disk_axis_ratio=True,
            disk_as_sersic=True,
            number_of_gaussians=2,
        )

        assert light.tag_beginner == "light__gaussians_x2"


class TestPipelineMassSettings:
    def test__no_shear_tag(self):
        mass = al.setup.Mass(no_shear=False)
        assert mass.no_shear_tag == "__with_shear"

        mass = al.setup.Mass(no_shear=True)
        assert mass.no_shear_tag == "__no_shear"

    def test__align_light_dark_tag(self):

        mass = al.setup.Mass(align_light_dark_centre=False)
        assert mass.align_light_dark_centre_tag == ""
        mass = al.setup.Mass(align_light_dark_centre=True)
        assert mass.align_light_dark_centre_tag == "__align_light_dark_centre"

    def test__align_bulge_dark_tag(self):
        mass = al.setup.Mass(align_bulge_dark_centre=False)
        assert mass.align_bulge_dark_centre_tag == ""
        mass = al.setup.Mass(align_bulge_dark_centre=True)
        assert mass.align_bulge_dark_centre_tag == "__align_bulge_dark_centre"

    def test__fix_lens_light_tag(self):
        mass = al.setup.Mass(fix_lens_light=False)
        assert mass.fix_lens_light_tag == ""
        mass = al.setup.Mass(fix_lens_light=True)
        assert mass.fix_lens_light_tag == "__fix_lens_light"

    def test__tag(self):

        mass = al.setup.Mass(
            no_shear=True, align_light_dark_centre=True, fix_lens_light=True
        )
        mass.type_tag = ""

        assert mass.tag == "mass____no_shear__align_light_dark_centre__fix_lens_light"

        mass = al.setup.Mass(align_bulge_dark_centre=True)

        mass.type_tag = "test"

        assert mass.tag == "mass__test__with_shear__align_bulge_dark_centre"

    def test__tag_beginner(self):

        mass = al.setup.Mass(
            no_shear=True, align_light_dark_centre=True, fix_lens_light=True
        )

        assert (
            mass.tag_beginner
            == "mass__no_shear__align_light_dark_centre__fix_lens_light"
        )

        mass = al.setup.Mass(align_bulge_dark_centre=True)

        assert mass.tag_beginner == "mass__with_shear__align_bulge_dark_centre"



import autofit as af
import autolens as al
from test_autolens.integration.tests.imaging import runner

test_type = "lens_only"
test_name = "lens_light"
data_type = "lens_light_dev_vaucouleurs"
data_resolution = "lsst"


def make_pipeline(name, phase_folders, optimizer_class=af.MultiNest):
    phase1 = al.PhaseImaging(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(lens=al.GalaxyModel(redshift=0.5, sersic=al.lp.EllipticalSersic)),
        optimizer_class=optimizer_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.8

    return al.PipelineDataset(name, phase1)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



from autolens.pipeline.phase import abstract


class Result(abstract.result.Result):
    @property
    def most_likely_fit(self):

        hyper_image_sky = self.analysis.hyper_image_sky_for_instance(
            instance=self.instance
        )

        hyper_background_noise = self.analysis.hyper_background_noise_for_instance(
            instance=self.instance
        )

        return self.analysis.masked_imaging_fit_for_tracer(
            tracer=self.most_likely_tracer,
            hyper_image_sky=hyper_image_sky,
            hyper_background_noise=hyper_background_noise,
        )

    @property
    def mask(self):
        return self.most_likely_fit.mask

    @property
    def positions(self):
        return self.most_likely_fit.masked_dataset.positions

    @property
    def pixelization(self):
        for galaxy in self.most_likely_fit.tracer.galaxies:
            if galaxy.pixelization is not None:
                return galaxy.pixelization

    @property
    def most_likely_pixelization_grids_of_planes(self):
        return self.most_likely_tracer.sparse_image_plane_grids_of_planes_from_grid(
            grid=self.most_likely_fit.grid
        )



from test import positions__offset_centre
from test_autolens.integration.tests.imaging.runner import run_a_mock


class TestCase:
    def _test_positions__offset_centre(self):
        run_a_mock(positions__offset_centre)



import autolens as al

import numpy as np
import time

repeats = 1

total_visibilities = 10000
pixelization_shape_2d = (30, 30)

real_space_shape_2d = (100, 100)
real_space_pixel_scales = 0.05
real_space_sub_size = 1
real_space_radius = 3.0

image_pixels = real_space_shape_2d[0] * real_space_shape_2d[1]
source_pixels = pixelization_shape_2d[0] * pixelization_shape_2d[1]

shape_data = 8 * total_visibilities
shape_preloads = total_visibilities * image_pixels * 2
shape_mapping_matrix = total_visibilities * source_pixels

total_shape = shape_data + shape_preloads + shape_mapping_matrix

print("Data Memory Use (GB) = " + str(shape_data * 8e-9))
print("PreLoad Memory Use (GB) = " + str(shape_preloads * 8e-9))
print("Mapping Matrix Memory Use (GB) = " + str(shape_mapping_matrix * 8e-9))
print("Total Memory Use (GB) = " + str(total_shape * 8e-9))
print()

# Only delete this if the memory use looks... Okay
# stop

visibilities = al.visibilities.ones(shape_1d=(total_visibilities,))
uv_wavelengths = np.ones(shape=(total_visibilities, 2))
noise_map = al.visibilities.ones(shape_1d=(total_visibilities,))

interferometer = al.interferometer(
    visibilities=visibilities, noise_map=noise_map, uv_wavelengths=uv_wavelengths
)

print("Real space sub grid size = " + str(real_space_sub_size))
print("Real space circular mask radius = " + str(real_space_radius) + "\n")
print("pixelization shape = " + str(pixelization_shape_2d) + "\n")

lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mp.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
    ),
)

pixelization = al.pix.VoronoiMagnification(shape=pixelization_shape_2d)

source_galaxy = al.Galaxy(
    redshift=1.0,
    pixelization=pixelization,
    regularization=al.reg.Constant(coefficient=1.0),
)

mask = al.mask.circular(
    shape_2d=real_space_shape_2d,
    pixel_scales=real_space_pixel_scales,
    sub_size=real_space_sub_size,
    radius=real_space_radius,
)

masked_interferometer = al.masked.interferometer(
    interferometer=interferometer, real_space_mask=mask
)

print("Number of points = " + str(masked_interferometer.grid.sub_shape_1d) + "\n")
print(
    "Number of visibilities = "
    + str(masked_interferometer.visibilities.shape_1d)
    + "\n"
)

start_overall = time.time()

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

start = time.time()
for i in range(repeats):
    traced_grid = tracer.traced_grids_of_planes_from_grid(
        grid=masked_interferometer.grid
    )[-1]
diff = time.time() - start
print("Time to Setup Traced Grid = {}".format(diff / repeats))

start = time.time()
for i in range(repeats):
    traced_sparse_grid = tracer.traced_sparse_grids_of_planes_from_grid(
        grid=masked_interferometer.grid
    )[-1]

diff = time.time() - start
print("Time to Setup Traced Sparse Grid = {}".format(diff / repeats))

start = time.time()
for i in range(repeats):
    mapper = pixelization.mapper_from_grid_and_sparse_grid(
        grid=traced_grid, sparse_grid=traced_sparse_grid, inversion_uses_border=True
    )
diff = time.time() - start
print("Time to create mapper (Border Relocation) = {}".format(diff / repeats))

start = time.time()
for i in range(repeats):
    mapping_matrix = mapper.mapping_matrix
diff = time.time() - start
print("Time to compute mapping matrix = {}".format(diff / repeats))

start = time.time()
for i in range(repeats):
    transformed_mapping_matrices = masked_interferometer.transformer.transformed_mapping_matrices_from_mapping_matrix(
        mapping_matrix=mapping_matrix
    )
diff = time.time() - start
print("Time to compute transformed mapping matrices = {}".format(diff / repeats))

start = time.time()
for i in range(repeats):
    real_data_vector = al.util.inversion.data_vector_from_transformed_mapping_matrix_and_data(
        transformed_mapping_matrix=transformed_mapping_matrices[0],
        visibilities=masked_interferometer.visibilities[:, 0],
        noise_map=masked_interferometer.noise_map[:, 0],
    )
diff = time.time() - start
print("Time to compute real data vector = {}".format(diff / repeats))


start = time.time()
for i in range(repeats):
    imag_data_vector = al.util.inversion.data_vector_from_transformed_mapping_matrix_and_data(
        transformed_mapping_matrix=transformed_mapping_matrices[1],
        visibilities=masked_interferometer.visibilities[:, 1],
        noise_map=masked_interferometer.noise_map[:, 1],
    )
diff = time.time() - start
print("Time to compute imaginary data vector = {}".format(diff / repeats))


start = time.time()
for i in range(repeats):
    real_curvature_matrix = al.util.inversion.curvature_matrix_from_transformed_mapping_matrix(
        transformed_mapping_matrix=transformed_mapping_matrices[0],
        noise_map=noise_map[:, 0],
    )
diff = time.time() - start
print("Time to compute real curvature matrix = {}".format(diff / repeats))

start = time.time()
for i in range(repeats):
    imag_curvature_matrix = al.util.inversion.curvature_matrix_from_transformed_mapping_matrix(
        transformed_mapping_matrix=transformed_mapping_matrices[1],
        noise_map=noise_map[:, 1],
    )
diff = time.time() - start
print("Time to compute imaginary curvature matrix = {}".format(diff / repeats))

start = time.time()
for i in range(repeats):
    regularization_matrix = al.util.regularization.constant_regularization_matrix_from_pixel_neighbors(
        coefficient=1.0,
        pixel_neighbors=mapper.pixelization_grid.pixel_neighbors,
        pixel_neighbors_size=mapper.pixelization_grid.pixel_neighbors_size,
    )
diff = time.time() - start
print("Time to compute reguarization matrix = {}".format(diff / repeats))

start = time.time()
for i in range(repeats):
    real_curvature_reg_matrix = np.add(real_curvature_matrix, regularization_matrix)
    imag_curvature_reg_matrix = np.add(imag_curvature_matrix, regularization_matrix)
    data_vector = np.add(real_data_vector, imag_data_vector)
    curvature_reg_matrix = np.add(real_curvature_reg_matrix, imag_curvature_reg_matrix)
diff = time.time() - start
print("Time to compute curvature reguarization Matrices = {}".format(diff / repeats))

start = time.time()
for i in range(repeats):
    reconstruction = np.linalg.solve(curvature_reg_matrix, data_vector)
diff = time.time() - start
print("Time to peform reconstruction = {}".format(diff / repeats))

start = time.time()
for i in range(repeats):
    real_mapped_visibilities = al.util.inversion.mapped_reconstructed_data_from_mapping_matrix_and_reconstruction(
        mapping_matrix=transformed_mapping_matrices[0], reconstruction=reconstruction
    )
diff = time.time() - start
print(
    "Time to compute mapped real visibility reconstruction = {}".format(diff / repeats)
)

start = time.time()
for i in range(repeats):
    imag_mapped_visibilities = al.util.inversion.mapped_reconstructed_data_from_mapping_matrix_and_reconstruction(
        mapping_matrix=transformed_mapping_matrices[1], reconstruction=reconstruction
    )
diff = time.time() - start
print(
    "Time to compute mapped imaginary visibility reconstruction = {}".format(
        diff / repeats
    )
)

start = time.time()
for i in range(repeats):
    al.fit(masked_dataset=masked_interferometer, tracer=tracer)
diff = time.time() - start
print("Time to perform complete fit = {}".format(diff / repeats))

print()



import builtins
import numpy as np
import pytest

import autofit as af
import autolens as al
from autofit import Paths
from autolens import exc


class MockAnalysis:
    def __init__(self, number_galaxies, shape, value):
        self.number_galaxies = number_galaxies
        self.shape = shape
        self.value = value

    # noinspection PyUnusedLocal
    def galaxy_images_for_model(self, model):
        return self.number_galaxies * [np.full(self.shape, self.value)]


class MockMask:
    pass


class Optimizer:
    def __init__(self, phase_name="dummy_phase"):
        self.phase_name = phase_name
        self.phase_path = ""


class DummyPhaseImaging(af.AbstractPhase):
    def make_result(self, result, analysis):
        pass

    def __init__(self, phase_name, phase_tag=""):
        super().__init__(Paths(phase_name=phase_name, phase_tag=phase_tag))
        self.dataset = None
        self.positions = None
        self.results = None
        self.mask = None

        self.optimizer = Optimizer(phase_name)

    def run(self, dataset, results, mask=None, positions=None):
        self.dataset = dataset
        self.results = results
        self.mask = mask
        self.positions = positions
        self.assert_and_save_pickle()
        return af.Result(af.ModelInstance(), 1)


class MockImagingData:
    pass


class MockFile:
    def __init__(self):
        self.text = None
        self.filename = None

    def write(self, text):
        self.text = text

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


@pytest.fixture(name="mock_files", autouse=True)
def make_mock_file(monkeypatch):
    files = []

    def mock_open(filename, flag, *args, **kwargs):
        assert flag in ("w+", "w+b", "a")
        file = MockFile()
        file.filename = filename
        files.append(file)
        return file

    monkeypatch.setattr(builtins, "open", mock_open)
    yield files


class TestMetaData:
    def test_files(self, mock_files):
        pipeline = al.PipelineDataset(
            "pipeline_name", DummyPhaseImaging(phase_name="phase_name")
        )
        pipeline.run(dataset=MockImagingData(), mask=MockMask(), data_name="data_name")

        assert (
            mock_files[1].text
            == "pipeline=pipeline_name\nphase=phase_name\ndataset_name=data_name\nphase_tag=\npipeline_tag=None"
        )

        assert "phase_name///optimizer.pickle" in mock_files[2].filename


class TestPassMask:
    def test_pass_mask(self):
        mask = MockMask()
        phase_1 = DummyPhaseImaging("one")
        phase_2 = DummyPhaseImaging("two")
        pipeline = al.PipelineDataset("", phase_1, phase_2)
        pipeline.run(dataset=MockImagingData(), mask=mask)

        assert phase_1.mask is mask
        assert phase_2.mask is mask


class TestPassPositions:
    def test_pass_positions(self):
        positions = [[(1.0, 1.0), (2.0, 2.0)]]
        phase_1 = DummyPhaseImaging("one")
        phase_2 = DummyPhaseImaging("two")
        pipeline = al.PipelineDataset("", phase_1, phase_2)
        pipeline.run(dataset=MockImagingData(), mask=MockMask(), positions=positions)

        assert phase_1.positions == positions
        assert phase_2.positions == positions


class TestPipelineImaging:
    def test_run_pipeline(self):
        phase_1 = DummyPhaseImaging("one")
        phase_2 = DummyPhaseImaging("two")

        pipeline = al.PipelineDataset("", phase_1, phase_2)

        pipeline.run(dataset=MockImagingData(), mask=MockMask())

        assert len(phase_2.results) == 2

    def test_addition(self):
        phase_1 = DummyPhaseImaging("one")
        phase_2 = DummyPhaseImaging("two")
        phase_3 = DummyPhaseImaging("three")

        pipeline1 = al.PipelineDataset("", phase_1, phase_2)
        pipeline2 = al.PipelineDataset("", phase_3)

        assert (phase_1, phase_2, phase_3) == (pipeline1 + pipeline2).phases


class DummyPhasePositions(af.AbstractPhase):
    def make_result(self, result, analysis):
        pass

    def __init__(self, phase_name):
        super().__init__(Paths(phase_name=phase_name, phase_tag=""))
        self.positions = None
        self.results = None
        self.pixel_scales = None
        self.optimizer = Optimizer(phase_name)

    def run(self, positions, pixel_scales, results):
        self.positions = positions
        self.pixel_scales = pixel_scales
        self.results = results
        return af.Result(af.ModelInstance(), 1)


class TestPipelinePositions:
    def test_run_pipeline(self):
        phase_1 = DummyPhasePositions(phase_name="one")
        phase_2 = DummyPhasePositions(phase_name="two")
        pipeline = al.PipelinePositions("", phase_1, phase_2)

        pipeline.run(positions=None, pixel_scales=None)

        assert len(phase_2.results) == 2

    def test_addition(self):
        phase_1 = DummyPhasePositions("one")
        phase_2 = DummyPhasePositions("two")
        phase_3 = DummyPhasePositions("three")

        pipeline1 = al.PipelinePositions("", phase_1, phase_2)
        pipeline2 = al.PipelinePositions("", phase_3)

        assert (phase_1, phase_2, phase_3) == (pipeline1 + pipeline2).phases



import autofit as af
from autoastro.galaxy import galaxy as g


class Result(af.Result):
    def __init__(
        self,
        instance,
        figure_of_merit,
        previous_model,
        gaussian_tuples,
        analysis,
        optimizer,
    ):
        """
        The result of a phase
        """
        super().__init__(
            instance=instance,
            figure_of_merit=figure_of_merit,
            previous_model=previous_model,
            gaussian_tuples=gaussian_tuples,
        )

        self.analysis = analysis
        self.optimizer = optimizer

    @property
    def most_likely_tracer(self):
        return self.analysis.tracer_for_instance(instance=self.instance)

    @property
    def path_galaxy_tuples(self) -> [(str, g.Galaxy)]:
        """
        Tuples associating the names of galaxies with instances from the best fit
        """
        return self.instance.path_instance_tuples_for_class(cls=g.Galaxy)



from autolens.model.profiles import light_profiles as lp

from autolens.model.profiles.plotters import profile_plotters

sersic = al.EllipticalSersic(
    centre=(0.0, 0.0),
    axis_ratio=0.9,
    phi=0.0,
    intensity=1.0,
    effective_radius=1.0,
    sersic_index=4.0,
)

aplt_array.profile.luminosity_within_circle_in_electrons_per_second_as_function_of_radius(
    light_profile=sersic, minimum_radius=1.0e-4, maximum_radius=10.0
)



import os

import autofit as af
import autolens as al

test_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))


def pixel_scale_from_data_resolution(data_resolution):
    """Determine the pixel scale from a data_type resolution type based on real observations.

    These options are representative of LSST, Euclid, HST, over-sampled HST and Adaptive Optics image.

    Parameters
    ----------
    data_resolution : str
        A string giving the resolution of the desired data_type type (LSST | Euclid | HST | HST_Up | AO).
    """
    if data_resolution == "lsst":
        return (0.2, 0.2)
    elif data_resolution == "euclid":
        return (0.1, 0.1)
    elif data_resolution == "hst":
        return (0.05, 0.05)
    elif data_resolution == "hst_up":
        return (0.03, 0.03)
    elif data_resolution == "ao":
        return (0.01, 0.01)
    else:
        raise ValueError(
            "An invalid data_type resolution was entered - ", data_resolution
        )


def load_test_imaging(
    data_type, data_resolution, psf_shape_2d=(11, 11), lens_name=None
):

    pixel_scales = pixel_scale_from_data_resolution(data_resolution=data_resolution)

    dataset_path = af.path_util.make_and_return_path_from_path_and_folder_names(
        path=test_path, folder_names=["dataset", "imaging", data_type, data_resolution]
    )

    return al.imaging.from_fits(
        image_path=dataset_path + "/image.fits",
        psf_path=dataset_path + "/psf.fits",
        noise_map_path=dataset_path + "/noise_map.fits",
        pixel_scales=pixel_scales,
        resized_psf_shape=psf_shape_2d,
        lens_name=lens_name,
    )



import autofit as af
from autoastro.galaxy import galaxy as g
from autolens.lens import ray_tracing


class Analysis(af.Analysis):
    def __init__(self, cosmology, results):

        self.cosmology = cosmology

        # TODO : This if loop is because of an OptimizerGridSeach, where the 'best_result' we do not want to update
        # TODO: the hyper images using.

        if results is not None and results.last is not None:

            if hasattr(results[-1], "hyper_galaxy_image_path_dict"):

                self.hyper_galaxy_image_path_dict = results[
                    -1
                ].hyper_galaxy_image_path_dict

                self.hyper_model_image = results[-1].hyper_model_image

            else:

                self.hyper_galaxy_image_path_dict = results[
                    -2
                ].hyper_galaxy_image_path_dict

                self.hyper_model_image = results[-2].hyper_model_image

    def hyper_image_sky_for_instance(self, instance):

        if hasattr(instance, "hyper_image_sky"):
            return instance.hyper_image_sky
        else:
            return None

    def hyper_background_noise_for_instance(self, instance):

        if hasattr(instance, "hyper_background_noise"):
            return instance.hyper_background_noise
        else:
            return None

    def tracer_for_instance(self, instance):
        return ray_tracing.Tracer.from_galaxies(
            galaxies=instance.galaxies, cosmology=self.cosmology
        )

    def associate_hyper_images(self, instance: af.ModelInstance) -> af.ModelInstance:
        """
        Takes images from the last result, if there is one, and associates them with galaxies in this phase
        where full-path galaxy names match.

        If the galaxy collection has a different name then an association is not made.

        e.g.
        galaxies.lens will match with:
            galaxies.lens
        but not with:
            galaxies.lens
            galaxies.source

        Parameters
        ----------
        instance
            A model instance with 0 or more galaxies in its tree

        Returns
        -------
        instance
           The input instance with images associated with galaxies where possible.
        """
        if hasattr(self, "hyper_galaxy_image_path_dict"):
            for galaxy_path, galaxy in instance.path_instance_tuples_for_class(
                g.Galaxy
            ):
                if galaxy_path in self.hyper_galaxy_image_path_dict:
                    galaxy.hyper_model_image = self.hyper_model_image
                    galaxy.hyper_galaxy_image = self.hyper_galaxy_image_path_dict[
                        galaxy_path
                    ]

        return instance



import autolens as al

from test_autolens.simulate.imaging import simulate_util

imaging = simulate_util.load_test_imaging(
    data_type="lens_sie__source_smooth__offset_centre", data_resolution="lsst"
)


def fit_with_offset_centre(centre):

    mask = al.mask.elliptical(
        shape_2d=imaging.shape_2d,
        pixel_scales=imaging.pixel_scales,
        major_axis_radius=3.0,
        axis_ratio=1.0,
        phi=0.0,
        centre=centre,
    )

    # The lines of code below do everything we're used to, that is, setup an image and its grid, mask it, trace it
    # via a tracer, setup the rectangular mapper, etc.
    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mp.EllipticalIsothermal(
            centre=(2.0, 2.0), einstein_radius=1.2, axis_ratio=0.7, phi=45.0
        ),
    )
    source_galaxy = al.Galaxy(
        redshift=1.0,
        pixelization=al.pix.VoronoiMagnification(shape=(20, 20)),
        regularization=al.reg.Constant(coefficient=1.0),
    )

    masked_imaging = al.masked.imaging(imaging=imaging, mask=mask)

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])
    fit = al.fit(masked_dataset=masked_imaging, tracer=tracer)

    return fit


fit = fit_with_offset_centre(centre=(2.0, 2.0))

# aplt.fit_imaging.subplot_for_plane(fit=fit, plane_index=1, plot_source_grid=True)

aplt.inversion.reconstruction(inversion=fit.inversion, include_grid=True)

stop

aplt.fit_imaging.subplot_fit_imaging(
    fit=fit,
    mask=True,
    positions=[[(2.2, 2.2), (-0.2, -0.2), (-0.2, 2.2), (2.2, -0.2)]],
    include_image_plane_pix=True,
)

fit = fit_with_offset_centre(centre=(2.05, 2.05))


aplt.fit_imaging.subplot_fit_imaging(
    fit=fit,
    mask=True,
    positions=[[(2.2, 2.2), (-0.2, -0.2), (-0.2, 2.2), (2.2, -0.2)]],
    include_image_plane_pix=True,
)

fit = fit_with_offset_centre(centre=(2.1, 2.1))

aplt.fit_imaging.subplot_fit_imaging(
    fit=fit,
    mask=True,
    positions=[[(2.2, 2.2), (-0.2, -0.2), (-0.2, 2.2), (2.2, -0.2)]],
    include_image_plane_pix=True,
)

fit = fit_with_offset_centre(centre=(2.95, 2.95))

aplt.fit_imaging.subplot_fit_imaging(
    fit=fit,
    mask=True,
    positions=[[(2.2, 2.2), (-0.2, -0.2), (-0.2, 2.2), (2.2, -0.2)]],
    include_image_plane_pix=True,
)

fit = fit_with_offset_centre(centre=(5.9, 5.9))

aplt.fit_imaging.subplot_fit_imaging(
    fit=fit,
    mask=True,
    positions=[[(2.2, 2.2), (-0.2, -0.2), (-0.2, 2.2), (2.2, -0.2)]],
    include_image_plane_pix=True,
)



from astropy.io import fits
import numpy as np
import os

path = "{}/../test_files/arrays/".format(os.path.dirname(os.path.realpath(__file__)))

array = np.array([[1.0, 1.0, 1.0], [1.0, 2.0, 1.0], [1.0, 1.0, 1.0]])

fits.writeto(data=array, filename=path + "3x3_ones_central_two.fits")
stop

array1 = np.ones((3, 3))
array2 = 2.0 * np.ones((3, 3))
array3 = 3.0 * np.ones((3, 3))
array4 = 4.0 * np.ones((3, 3))
array5 = 5.0 * np.ones((3, 3))
array6 = 6.0 * np.ones((3, 3))
array7 = 7.0 * np.ones((3, 3))
array8 = 8.0 * np.ones((3, 3))

fits.writeto(data=array2, filename=path + "3x3_twos.fits")
fits.writeto(data=array3, filename=path + "3x3_threes.fits")
fits.writeto(data=array4, filename=path + "3x3_fours.fits")
fits.writeto(data=array5, filename=path + "3x3_fives.fits")
fits.writeto(data=array6, filename=path + "3x3_sixes.fits")
fits.writeto(data=array7, filename=path + "3x3_sevens.fits")
fits.writeto(data=array8, filename=path + "3x3_eights.fits")
stop
new_hdul = fits.HDUList()
new_hdul.append(fits.ImageHDU(array1))
new_hdul.append(fits.ImageHDU(array2))
new_hdul.append(fits.ImageHDU(array3))
new_hdul.append(fits.ImageHDU(array4))
new_hdul.append(fits.ImageHDU(array5))
new_hdul.append(fits.ImageHDU(array6))
new_hdul.append(fits.ImageHDU(array7))
new_hdul.append(fits.ImageHDU(array8))

new_hdul.writeto(path + "3x3_multiple_hdu.fits")



import autofit as af
import autolens as al
from test_autolens.integration.tests.imaging import runner

test_type = "lens__source"
test_name = "lens_mass__source_x2__hyper"
data_type = "lens_sie__source_smooth"
data_resolution = "lsst"


def make_pipeline(name, phase_folders, optimizer_class=af.MultiNest):

    phase1 = al.PhaseImaging(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(redshift=0.5, mass=al.mp.EllipticalIsothermal),
            source_0=al.GalaxyModel(redshift=1.0, light=al.lp.EllipticalSersic),
        ),
        optimizer_class=optimizer_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 60
    phase1.optimizer.sampling_efficiency = 0.7

    phase2 = al.PhaseImaging(
        phase_name="phase_2",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=phase1.result.model.galaxies.lens.mass
            ),
            source_0=al.GalaxyModel(
                redshift=1.0, light=phase1.result.model.galaxies.source_0.light
            ),
            source_1=al.GalaxyModel(redshift=1.0, light=al.lp.EllipticalSersic),
        ),
        optimizer_class=optimizer_class,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 60
    phase2.optimizer.sampling_efficiency = 0.7

    phase2 = phase2.extend_with_multiple_hyper_phases(hyper_galaxy=True)

    class HyperLensSourcePlanePhase(al.PhaseImaging):
        def customize_priors(self, results):

            self.galaxies.source_0.hyper_galaxy = (
                results.last.hyper_combined.instance.galaxies.source_0.hyper_galaxy
            )

            self.galaxies.source_1.hyper_galaxy = (
                results.last.hyper_combined.instance.galaxies.source_1.hyper_galaxy
            )

    phase3 = HyperLensSourcePlanePhase(
        phase_name="phase_3",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=phase2.result.model.galaxies.lens.mass
            ),
            source_0=al.GalaxyModel(
                redshift=1.0,
                light=phase2.result.model.galaxies.source_0.light,
                hyper_galaxy=phase2.result.hyper_combined.instance.galaxies.source_0.hyper_galaxy,
            ),
            source_1=al.GalaxyModel(
                redshift=1.0,
                light=phase2.result.model.galaxies.source_1.light,
                hyper_galaxy=phase2.result.hyper_combined.instance.galaxies.source_1.hyper_galaxy,
            ),
        ),
        optimizer_class=optimizer_class,
    )

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 40
    phase3.optimizer.sampling_efficiency = 0.8

    return al.PipelineDataset(name, phase1, phase2, phase3)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])
